import 'package:data_db/data_db.dart';
import 'package:static_postgres_orm/static_postgres_orm.dart';
import 'pg_table_generator.dart';
import 'package:postgres/postgres.dart';

///Looad a connection to Postgres Database and automatically generates
///ORM for all tables of target database
///
///Include "path" named parameter to specify the local pasth to generateds files
///
///Include "schemaInName" named parameter if you want supress the schame name of
///table in generateds files and class names of thats files
Future<bool> generate_pg_orm(PostgreSQLConnection connection,
    {String? path, bool schemaInName = true}) async {
  try {
    await _generateOrm(connection, (await _getTables(connection)), schemaInName,
        path: path);

    return true;
  } finally {}
}

Future<void> _generateOrm(
    PostgreSQLConnection connection, Dataset tables, bool schemaInName,
    {String? path}) async {
  tables.forEach((record) {
    record.valueAsString('table_schema').map((table_schema) {
      record.valueAsString('table_name').map((table_name) async {
        gernerateTableORM(
            await _getFieldDescription(connection, table_schema, table_name),
            schemaInName,
            path: path);
      });
    });
  });
}

extension _StringList on List<String> {
  String asText() {
    var x = '';
    forEach((element) {
      x += '\n' + element;
    });
    return x.replaceFirst('\n', '');
  }
}

PostgreSQLConnection _newConnection(PostgreSQLConnection connection) {
  return PostgreSQLConnection(
      connection.host, connection.port, connection.databaseName,
      username: connection.username, password: connection.password);
}

Future<Dataset> _getTables(PostgreSQLConnection connection) async {
  var sql = <String>[];

  sql
    ..add('select')
    ..add('cast (table_schema as varchar),')
    ..add('cast(table_name as varchar)')
    ..add('from information_schema.tables')
    ..add("where table_schema not in ('pg_catalog','information_schema')");
  var con = _newConnection(connection);
  await con.open();
  var r = pgSqlToDataset(await con.query(sql.asText()));
  await con.close();
  return r;
}

Future<Dataset> _getFieldDescription(PostgreSQLConnection connection,
    String schemaName, String tableName) async {
  var sql = <String>[];
  sql
    ..add(
        'SELECT columns.table_schema, columns.table_name, columns.ordinal_position,')
    ..add(
        'columns.column_name, cast(pg_type.oid as int) as type_id, columns.data_type,')
    ..add(
        "columns.column_default, case when LOWER(columns.is_nullable) = 'yes' then")
    ..add(
        'true else false end as is_nullable, columns.character_maximum_length,')
    ..add(
        'columns.numeric_precision, columns.numeric_scale, columns.datetime_precision,')
    ..add(
        'case when ( select pgc.conname from pg_constraint pgc join pg_namespace nsp on nsp.oid')
    ..add(
        '= pgc.connamespace join pg_class cls on pgc.conrelid = cls.oid left join ')
    ..add('information_schema.constraint_column_usage ccu on pgc.conname =')
    ..add(
        'ccu.constraint_name and nsp.nspname = ccu.constraint_schema where ccu.table_schema =')
    ..add(
        'columns.table_schema and ccu.table_name = columns.table_name and ccu.column_name = ')
    ..add(
        "columns.column_name and pgc.contype = 'p' ) <> '' then true else false")
    ..add(
        'end as is_pk FROM information_schema.columns inner join pg_catalog.pg_type on columns.udt_name =')
    ..add("pg_type.typname WHERE table_schema = '$schemaName'")
    ..add("and table_name = '$tableName'")
    ..add('order by 1,2,3');

  var con = _newConnection(connection);
  await con.open();
  var x = pgSqlToDataset(await con.query(sql.asText()));
  await con.close();
  return x;
}
